// QUBIO.cpp 
// Main window construction and application setup 
//
#include "stdafx.h"
#include "time.h"
#include "math.h"
#include <string>
using std::string;

#include "pdf.h"

const int IDM_FIRSTCHILD=1000;

//-- Forward declarations
LRESULT CALLBACK About(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK MainWndProc(HWND, UINT, WPARAM, LPARAM);

//-- special window - MDI frame client 
HWND g_hClientWin;				

//-- Array of hardware configuations
CConfigurations g_oCfgs;						// externed

const char * g_szFrameClass = "QUBIOMainWindow" ;
const char * g_szChildClass = "QUBIODialogWindow" ;
const char * g_szTitle = "QUB Input/Output (Acquisition/Generation)";

//----- bitmaps for buttons
HBITMAP g_hbUpdate[2];
HBITMAP g_hbPlay[3];
HBITMAP g_hbSynch[2];
HBITMAP g_hbStop[3];
HBITMAP g_hbLoop[3];
HBITMAP g_hbRecord[2];
HBITMAP g_hbDraw[3];
HBITMAP g_hbScan[2];
HBITMAP g_hbTimeSep[2];
HBITMAP g_hbPage;
HFONT g_fntSS8;

//----- Global Handle 
HMENU g_hMenuInit;

//----- WinMain app entry point and main message pump
int APIENTRY _tWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPTSTR lpCmdLine, int nCmdShow ) {
 	// Set global instance variable
	SetInstance( hInstance );
 	
 	// Register Window Classes 
	CreateWindowClass( g_szFrameClass, (WNDPROC)MainWndProc, hInstance, (LPCTSTR)IDI_QUBIO, (LPCTSTR)IDI_SMALL, 0, COLOR_APPWORKSPACE );
	CreateWindowClass( g_szChildClass, (WNDPROC)ChildWndProc, hInstance, (LPCTSTR)IDI_QUBIO, (LPCTSTR)IDI_SMALL, sizeof(HANDLE), COLOR_BTNFACE );
	RegisterReportClass( );
	
	//----- Initialization 
	// create and display the main program window.
	g_hMenuInit = LoadMenu( hInstance, (LPCTSTR)IDC_QUBIO );
	HACCEL hAccelTable = LoadAccelerators(hInstance, (LPCTSTR)IDC_QUBIO);

	g_hbUpdate[0] = LoadBitmap( hInstance, MAKEINTRESOURCE(IDB_UPDATE) );
	g_hbUpdate[1] = LoadBitmap( hInstance, MAKEINTRESOURCE(IDB_UPDATE1) );
	
	g_hbPlay[0] = LoadBitmap( hInstance, MAKEINTRESOURCE(IDB_PLAY) );
	g_hbPlay[1] = LoadBitmap( hInstance, MAKEINTRESOURCE(IDB_PLAY1) );
	g_hbPlay[2] = LoadBitmap( hInstance, MAKEINTRESOURCE(IDB_PLAY2) );

	g_hbSynch[0] = LoadBitmap( hInstance, MAKEINTRESOURCE(IDB_SYNCH) );
	g_hbSynch[1] = LoadBitmap( hInstance, MAKEINTRESOURCE(IDB_SYNCH1) );

	g_hbLoop[0] = LoadBitmap( hInstance, MAKEINTRESOURCE(IDB_LOOP) );
	g_hbLoop[1] = LoadBitmap( hInstance, MAKEINTRESOURCE(IDB_LOOP1) );
	g_hbLoop[2] = LoadBitmap( hInstance, MAKEINTRESOURCE(IDB_LOOP2) );

	g_hbStop[0] = LoadBitmap( hInstance, MAKEINTRESOURCE(IDB_STOP) );
	g_hbStop[1] = LoadBitmap( hInstance, MAKEINTRESOURCE(IDB_STOP1) );
	g_hbStop[2] = LoadBitmap( hInstance, MAKEINTRESOURCE(IDB_STOP2) );

	g_hbRecord[0] = LoadBitmap( hInstance, MAKEINTRESOURCE(IDB_RECORD) );
	g_hbRecord[1] = LoadBitmap( hInstance, MAKEINTRESOURCE(IDB_RECORD1) );
	
	g_hbScan[0] = LoadBitmap( hInstance, MAKEINTRESOURCE(IDB_SCAN1) );
	g_hbScan[1] = LoadBitmap( hInstance, MAKEINTRESOURCE(IDB_SCAN2) );
	
	g_hbDraw[0] = LoadBitmap( hInstance, MAKEINTRESOURCE(IDB_DRAW1) );
	g_hbDraw[1] = LoadBitmap( hInstance, MAKEINTRESOURCE(IDB_DRAW2) );
	g_hbDraw[2] = LoadBitmap( hInstance, MAKEINTRESOURCE(IDB_DRAW3) );
	
	g_hbTimeSep[0] = LoadBitmap( hInstance, MAKEINTRESOURCE(IDB_TIMESEP) );
	g_hbTimeSep[1] = LoadBitmap( hInstance, MAKEINTRESOURCE(IDB_TIMESEP1) );
	
	g_hbPage = LoadBitmap( hInstance, MAKEINTRESOURCE(IDB_PAGE) );

	g_fntSS8 = CreateFont( 15, 0, 0, 0, FW_REGULAR, 0, 0, 0, 
								DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, 
								PROOF_QUALITY /*DEFAULT_QUALITY*/, 
								FF_SWISS /* FF_ROMAN */ | VARIABLE_PITCH, NULL );
								
	HWND hFrameWin = CreateWindow( g_szFrameClass, g_szTitle, 
							WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN, 
							CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, 
							NULL, g_hMenuInit,  hInstance, NULL);
	if( ! hFrameWin )
		return FALSE;
								
	g_hClientWin = GetWindow( hFrameWin, GW_CHILD );	//ShowWindow( hFrameWin, nCmdShow );
	ShowWindow( hFrameWin, SW_MAXIMIZE );
	UpdateWindow( hFrameWin );

	//-- 
	InitCommonControls();

	//-- Data initialization.  Refresh menus after load - some menu items change.
	g_oCfgs.LoadFile();			
	OnGainTableUpdated( true );
	OnHardwareConfigUpdated( true );
	OnFeedbackUpdated( true );
	
	//-- if configuration file does not exist, open the hardware config window 
	if( ! g_oCfgs.ConfigFileExists( ) )
		PostMessage( hFrameWin, WM_COMMAND, ID_HARDWARECONFIG_NEW, 0);
	else 
		PostMessage( hFrameWin, WM_COMMAND, ID_VIEW_ERRORWINDOW, 0);
	
	//-- seed random number generator.
	srand( (unsigned)time( NULL ) );

	//TestPDFRpt();

	//-- Main message loop:
	MSG msg;
	while( GetMessage(&msg, NULL, 0, 0) ) {
		if( ! WindowListProcessMessage( &msg ) 
			&& ! TranslateMDISysAccel( g_hClientWin, &msg ) 
			&& ! TranslateAccelerator( hFrameWin, hAccelTable, &msg)
			) {	//msg.hwnd
			TranslateMessage(&msg);
			DispatchMessage(&msg);
			}
		}

	DeleteObject( g_fntSS8 );			

	DeleteObject( g_hbUpdate[0] );
	DeleteObject( g_hbUpdate[1] );
	DeleteObject( g_hbPlay[0] );
	DeleteObject( g_hbPlay[1] );
	DeleteObject( g_hbPlay[2] );
	DeleteObject( g_hbSynch[0] );
	DeleteObject( g_hbSynch[1] );
	DeleteObject( g_hbLoop[0] );
	DeleteObject( g_hbLoop[1] );
	DeleteObject( g_hbLoop[2] );
	DeleteObject( g_hbStop[0] );
	DeleteObject( g_hbStop[1] );
	DeleteObject( g_hbStop[2] );
	DeleteObject( g_hbRecord[0] );
	DeleteObject( g_hbRecord[1] );
	DeleteObject( g_hbScan[0] );
	DeleteObject( g_hbScan[1] );
	DeleteObject( g_hbDraw[0] );
	DeleteObject( g_hbDraw[1] );
	DeleteObject( g_hbDraw[2] );
	DeleteObject( g_hbTimeSep[0] );
	DeleteObject( g_hbTimeSep[1] );
	DeleteObject( g_hbPage );

	return (int) msg.wParam;
	}

//-----------------------------------------------------------------------------	
//  MainWndProc() -  Processes messages for the main window.
LRESULT CALLBACK MainWndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam){
	int wmId, wmEvent;
	HWND hChildWin, hDlg; 

	switch (message) {
		case WM_CREATE : {
			//---- Create the client of the frame - an MDI oddity 
			CLIENTCREATESTRUCT ccs;
			ccs.hWindowMenu=GetSubMenu( GetMenu(hWnd), 1/*position of menu*/ );
			ccs.idFirstChild=IDM_FIRSTCHILD;
			g_hClientWin = CreateWindow( "MDICLIENT", NULL, 
								WS_CHILD|WS_CLIPCHILDREN| WS_VISIBLE, 
								0, 0, 0, 0, hWnd, 
								(HMENU) 0xCAC, //  GetSubMenu(GetMenu(hwnd), WINDOWMENU), 
								GetInstance(), &ccs );
							
			NIInit();
		
			//---- create the report window 
			//CreateReportWindow();
			AddMessage( REPTYPE_NOTE, "QUBIO_STARTED" );
			}			
			return 0;

		case WM_ENTERMENULOOP:
			OnQuickSessionsUpdated();
			return 0;
 
		case WM_COMMAND:
			//----- Parse the menu selections:
			wmId    = LOWORD(wParam); 
			wmEvent = HIWORD(wParam);
			
			//-- Gain Table Edit - dynamic menu 
			if( wmId >= IDM_GAINTABLE && wmId < IDM_GAINTABLE+g_oCfgs.m_aGainTables.Count() )  {
				int iGainTable = wmId-IDM_GAINTABLE;
				CGainTable * pGainTable = g_oCfgs.GainTable(iGainTable);				
				INT_PTR ip = DialogBoxParam( GetInstance(), 
								MAKEINTRESOURCE(IDD_DLGGAINTABLE), 
								hWnd,
								(DLGPROC)GainTableDlgProc, 
								(LPARAM)pGainTable );
				if( ip==IDYES ) {		//-- deleted and confirmed
					g_oCfgs.m_aGainTables.DeleteItem( iGainTable );
					OnGainTableUpdated(  );
					}
				else if( ip==IDOK ) 
					OnGainTableUpdated(  );
				}
			
			//-- Hardware config Edit 
			if( wmId >= IDM_HARDWARECONFIG && wmId < IDM_HARDWARECONFIG+g_oCfgs.m_aConfigs.Count() )  {
				int iHdwCfg = wmId-IDM_HARDWARECONFIG;
				CConfiguration * pConfig = g_oCfgs.Config(iHdwCfg);				
				SendMessage( OpenConfigView( pConfig->m_cConfigName.ConstBuffer() ), WM_CONFIG_EDIT, (WPARAM)iHdwCfg, (LPARAM)pConfig );
				}

			//-- Feedback config Edit 
			if( wmId >= IDM_FEEDBACKCONFIG && wmId < IDM_FEEDBACKCONFIG+g_oCfgs.m_aFeedbackConfigs.Count() )  {
				int iFeedbackCfg = wmId-IDM_FEEDBACKCONFIG;
				CFeedback * pFeedback = g_oCfgs.FeedbackConfig(iFeedbackCfg);				
				INT_PTR ip = DialogBoxParam( GetInstance(), 
								MAKEINTRESOURCE(IDD_DLGFEEDBACK), 
								hWnd,
								(DLGPROC)FeedbackDlgProc, 
								(LPARAM)pFeedback );
				if( ip==IDYES ) {		//-- deleted and confirmed
					g_oCfgs.m_aFeedbackConfigs.DeleteItem( iFeedbackCfg );
					OnFeedbackUpdated(  );
					}
				else if( ip==IDOK ) 
					OnFeedbackUpdated(  );
				}

			// session quick-start
			if( wmId >= IDM_QUICKSESSION && wmId < IDM_QUICKSESSION+g_oCfgs.m_oSettings.m_aQuickSessions.Count() )  {
				string quickName = g_oCfgs.m_oSettings.m_aQuickSessions.GetAt(wmId - IDM_QUICKSESSION).Buffer();
				SendMessage( OpenSessionView( quickName.c_str() ), WM_SESSION_LOAD, 1, (LPARAM)quickName.c_str() );
				g_oCfgs.m_oSettings.addSession( quickName.c_str() );
				return 0;
			}

			switch (wmId) {
				//----- Open the Session window 	
				case ID_SESSION_NEW :
					SendMessage( OpenSessionView( "QUBIO Session" ), WM_SESSION_NEW, 0, 0 ); 
					return 0; 

				case ID_SESSION_OPEN : {
					//-- select the file to load 
					char achLongName[_MAX_PATH]={0};
					char achFileName[_MAX_PATH]={0};
					if( ! GetOpenFileName( achLongName, 
									"Session Files (*.qsn)\0*.qsn\0All Files (*.*)\0*.*\0\0", 
									"qsn", GetParent(hWnd), achFileName,
									& g_oCfgs.m_oSettings.m_cSessionPath ) ) 
						return 0;
									  
					//-- Open the session window
					SendMessage( OpenSessionView( achFileName ), WM_SESSION_LOAD, 0, (LPARAM)achLongName );
					g_oCfgs.m_oSettings.addSession( achLongName );
					}
					return 0; 
		
				//----- Open the Protocol window 	
				case ID_PROTOCOL_NEW :
					//-- Open the Protocol editor
					SendMessage( OpenProtocolView("QUBIO Protocol"), WM_PRO_NEW, 0, 0 );
					return 0; 

				case ID_PROTOCOL_OPEN : {
					//-- select the file to load 
					static char achLongName[_MAX_PATH]={0};
					static char achFileName[_MAX_PATH]={0};
					if( ! GetOpenFileName( achLongName, 
									"Protocol Files (*.qpr)\0*.qpr\0All Files (*.*)\0*.*\0\0", 
									"qpr", GetParent(hWnd), achFileName,
									& g_oCfgs.m_oSettings.m_cProtoPath  ) ) 
						return 0;
									  
					//-- Open the Protocol editor
					SendMessage( OpenProtocolView(achFileName), WM_PRO_LOAD, 0, (LPARAM)achLongName );
					}
					return 0; 
					
				//----- Open the Protocol window 	
				case ID_DIGITALPROTOCOL_NEW :
					SendMessage( OpenDigitalProtocolView("QUBIO Digital Protocol"), WM_DIGPRO_NEW, 0, 0 );
					return 0; 

				case ID_DIGITALPROTOCOL_OPEN : {
					//-- select the file to load 
					char achLongName[_MAX_PATH]={0};
					char achFileName[_MAX_PATH]={0};
					if( ! GetOpenFileName( achLongName, 
									"Digital Protocol Files (*.qdp)\0*.qdp\0All Files (*.*)\0*.*\0\0", 
									"qdp", GetParent(hWnd), achFileName,
									& g_oCfgs.m_oSettings.m_cDigiProtoPath  ) ) 
						return 0;
									  
					//-- Open the Protocol editor
					SendMessage( OpenDigitalProtocolView(achFileName), WM_DIGPRO_LOAD, 0, (LPARAM)achLongName );
					}
					return 0; 

				//----- Hardware config New
				case ID_HARDWARECONFIG_NEW :
					SendMessage( OpenConfigView( "New Configuration" ), WM_CONFIG_NEW, 0, 0 );
					return 0; 

				//---- Gain Tables 
				case ID_GAINTABLES_NEW :  {
					CGainTable * pGainTable = new CGainTable;
					INT_PTR ip = DialogBoxParam( GetInstance(), 
									MAKEINTRESOURCE(IDD_DLGGAINTABLE), 
									hWnd,
								   (DLGPROC)GainTableDlgProc, 
								   (LPARAM)pGainTable );
					if( ip==IDOK ) {
						pGainTable->m_bNew=false;
						g_oCfgs.m_aGainTables.AddItem( pGainTable );
						OnGainTableUpdated(  );
						}
					else 
						delete pGainTable;
					}
					return 0; 
					
				//---- Feedback configurations
				case ID_FEEDBACKCONFIGS_NEW :  {
					CFeedback * pFeedbackConfig = new CPID;
					INT_PTR ip = DialogBoxParam( GetInstance(), 
									MAKEINTRESOURCE(IDD_DLGFEEDBACK), 
									hWnd,
								   (DLGPROC)FeedbackDlgProc, 
								   (LPARAM)pFeedbackConfig );
					if( ip==IDOK ) {
						g_oCfgs.m_aFeedbackConfigs.AddItem( pFeedbackConfig );
						OnFeedbackUpdated(  );
						}
					else 
						delete pFeedbackConfig;
					}
					return 0; 
					
				//----- Display the error window 
				case ID_VIEW_ERRORWINDOW : 
					CreateReportWindow( g_hClientWin );
					break;

				//----- Open the settings window 	
				case IDM_GENPULSESEQ :
					OpenVwGenPulseProtocol( hWnd );
					return 0; 

				//----- Open the settings window 	
				case IDM_FILE_DEFAULTSETTINGS :
					DialogBoxParam( GetInstance(), MAKEINTRESOURCE(IDD_DLGQUBIOSETTINGS), hWnd,
								   (DLGPROC)QUBIOSettingsDlgProc, (LPARAM)&g_oCfgs.m_oSettings );
					return 0; 

				//----- Exit ... 
				case IDM_EXIT:
		            SendMessage (hWnd, WM_CLOSE, 0, 0) ;
					break;
                
                case ID_WINDOW_CLOSE : {		// Close the active window
					HWND h = (HWND) SendMessage (g_hClientWin, WM_MDIGETACTIVE, 0, 0) ;
					if( SendMessage( h, WM_QUERYENDSESSION, 0, 0) )
						SendMessage( g_hClientWin, WM_MDIDESTROY, (WPARAM) h, 0) ;
					}
					return 0 ;
					
				//----- Open the About Dialogue - Modal 
				case IDM_ABOUT:
					DialogBox( GetInstance(), (LPCTSTR)IDD_ABOUTBOX, hWnd, (DLGPROC)About);
					break;
					
				default : {            
					HWND h = (HWND) SendMessage (g_hClientWin, WM_MDIGETACTIVE, 0, 0) ;
					if( IsWindow (h) )	// Pass to active child...
						SendMessage( h, WM_COMMAND, wParam, lParam );
					}
					break ;        // ...and then to DefFrameProc
				}
				break;

		case WM_ACQNOTE_NEW : {
			CSession * pSes = (CSession *) wParam;
			
			//-- open the note dialog with a new note at current time 
			hChildWin = CreateMDIWindow( g_szChildClass, "Session Note", NULL, 
							CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, 
							g_hClientWin, GetInstance(), 0 );
			WindowListAdd( hChildWin, hDlg = CreateDialog( GetInstance(), MAKEINTRESOURCE(IDD_ACQNOTE), hChildWin, AcqNoteDlgProc ) );
			AcqNoteDlgProc( hDlg, WM_ACQNOTESETUP, (WPARAM)pSes, NULL );
			AcqNoteDlgProc( hDlg, WM_ACQNOTESETUP2, (WPARAM) g_hClientWin, lParam );
			}
			return 0; 

		case WM_DESTROY:
			g_oCfgs.SaveFile( );
			PostQuitMessage(0);
			return 0;

		}
	return DefFrameProc( hWnd, g_hClientWin, message, wParam, lParam);
	}

//=============================================================================
HWND OpenDlgWindow( const char * pchWindowName, int iID, DLGPROC lpDialogFunc, DWORD dwStyle=0 ) { 
	//-- Open the data window 
	HWND hChildWin = CreateMDIWindow( g_szChildClass, pchWindowName, dwStyle, 
							CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, 
							g_hClientWin, GetInstance(), 0 );
							
	HWND hDlg = CreateDialog( GetInstance(), MAKEINTRESOURCE(iID), hChildWin, lpDialogFunc );							
	WindowListAdd( hChildWin, hDlg );
	return hDlg;
	}

HWND OpenConfigView( const char * pchWindowName ) { 
	return OpenDlgWindow( pchWindowName, IDD_CONFIGVIEW, ConfigDlgProc );
	}

HWND OpenSessionView( const char * pchWindowName ) { 
	return OpenDlgWindow( pchWindowName, IDD_SESSIONVIEW, SessionDlgProc, WS_MAXIMIZE );
	}

HWND OpenProtocolView( const char * pchWindowName ) { 
	return OpenDlgWindow( pchWindowName, IDD_PROTOCOL, ProtocolDlgProc );
	}
	
HWND OpenDigitalProtocolView( const char * pchWindowName ) { 
	return OpenDlgWindow( pchWindowName, IDD_DIGITALPROTOCOL, DigitalProtocolDlgProc );
	}

//=============================================================================
//---- Clear and refill the gain table menu 
void OnGainTableUpdated( bool bNoSave ) {
	HMENU hMenu = GetSubMenu( GetSubMenu( g_hMenuInit, 0 ), 2 );		
	int i=0;
	while( 0!=DeleteMenu( hMenu, 2, MF_BYPOSITION ) && i<100 )
		++i;
	
	for( i=0; i<g_oCfgs.m_aGainTables.Count(); ++i )
		AppendMenu( hMenu, MF_BYPOSITION|MF_STRING, (UINT_PTR)(IDM_GAINTABLE+i), 
				g_oCfgs.GainTable(i)->m_cTableName.ConstBuffer() ); 

	if( ! bNoSave ) 
		g_oCfgs.SaveFile();
	}

//---- Clear and refill the Hardware configurations menu 
void OnHardwareConfigUpdated( bool bNoSave ) {
	HMENU hMenu = GetSubMenu( GetSubMenu( g_hMenuInit, 0 ), 3 );		
	int i=0;
	while( 0!=DeleteMenu( hMenu, 2, MF_BYPOSITION ) && i<100 )
		++i;
	
	for( i=0; i<g_oCfgs.m_aConfigs.Count(); ++i )
		AppendMenu( hMenu, MF_BYPOSITION|MF_STRING, (UINT_PTR)(IDM_HARDWARECONFIG+i), 
				g_oCfgs.Config(i)->m_cConfigName.ConstBuffer() ); 

	if( ! bNoSave ) 
		g_oCfgs.SaveFile();
	}

//---- Clear and refill the Hardware configurations menu 
void OnFeedbackUpdated( bool bNoSave ) {
	HMENU hMenu = GetSubMenu( GetSubMenu( g_hMenuInit, 0 ), 4 );		
	int i=0;
	while( 0!=DeleteMenu( hMenu, 2, MF_BYPOSITION ) && i<100 )
		++i;
	
	for( i=0; i<g_oCfgs.m_aFeedbackConfigs.Count(); ++i )
		AppendMenu( hMenu, MF_BYPOSITION|MF_STRING, (UINT_PTR)(IDM_FEEDBACKCONFIG+i), 
				g_oCfgs.FeedbackConfig(i)->m_cName.ConstBuffer() ); 

	if( ! bNoSave ) 
		g_oCfgs.SaveFile();
	}

void OnQuickSessionsUpdated() {
	HMENU hMenu = GetSubMenu( GetSubMenu( g_hMenuInit, 0 ), 0 );
	int i=0; 
	while ( 0!=DeleteMenu( hMenu, 3, MF_BYPOSITION ) && i<100 )
		++i;

	for( i=0; i<g_oCfgs.m_oSettings.m_aQuickSessions.Count(); ++i )
		AppendMenu( hMenu, MF_BYPOSITION|MF_STRING, (UINT_PTR)(IDM_QUICKSESSION+i),
					g_oCfgs.m_oSettings.m_aQuickSessions[i].ConstBuffer() );
}

//=============================================================================
//---- Message handler for about box.
LRESULT CALLBACK About(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam){
	switch (message){
	case WM_INITDIALOG:
		return TRUE;

	case WM_COMMAND:
		if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL) {
			EndDialog(hDlg, LOWORD(wParam));
			return TRUE;
			}
		break;
		}
	return FALSE;
	}
